#include <debugging_regexis024/debug_through_graphviz.h>
#include <unistd.h>
#include <fcntl.h>
#include <wait.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <algorithm>
#include <libregexis024vm/vm_opcodes.h>
#include <libregexis024fa/tracking_fa_nodes.h>

namespace regexis024 {
    const char* one_char_read_color = "black";
    const char* forking_color = "darkorchid1";
    const char* look_one_behind_color = "darkslateblue";
    const char* look_one_ahead_color = "coral1";
    const char* track_array_mov_imm_color = "lightblue2";
    const char* track_array_mov_halfinvariant_color = "lightseagreen";
    const char* match_pending_lob_color = "darkgoldenrod2";
    const char* match_color = "gold";
    const char* det_char_crossroads_color = "navy";
    const char* error_color = "crimson";
    const char* STAR = "★";

    const char* get_associated_color(FA_Node* node){
        switch (node->type) {
#define ccase(tn) case tn: return tn##_color;
            ccase(one_char_read)
            ccase(forking)
            ccase(look_one_behind)
            ccase(look_one_ahead)
            ccase(track_array_mov_imm)
            ccase(track_array_mov_halfinvariant)
            ccase(det_char_crossroads)
            case match:
                return dynamic_cast<FA_NodeOfMatch*>(node)->ext_filter_added ? match_pending_lob_color : match_color;
            default:
                return "black";
#undef ccase
        }
    }

    struct NodesProblems{
        size_t actual_refcount = 0;
        bool refcount_problem = false;
        size_t edges_point_to_null = 0;
    };

    struct EdgesProblems {
        bool points_to_null = false;
        explicit EdgesProblems(bool points_to_null): points_to_null(points_to_null) {}
    };

    std::string get_applied_edge_attributes(FA_Node* node, const NodesProblems& np, const EdgesProblems& ep){
        std::string res = "color=";
        if (ep.points_to_null) {
            res += error_color;
        } else {
            res += get_associated_color(node);
            if (node->type == one_char_read || node->type == det_char_crossroads)
                res += " style=bold";
        }
        return res;
    }

    std::string get_applied_node_attributes(FA_Node* node, const NodesProblems& bd){
        std::string res = "color=";
        res += get_associated_color(node);
        if (bd.refcount_problem)
            res += " fontcolor=crimson";
        if ((node->type == match) ||
                 (node->type == det_char_crossroads && dynamic_cast<FA_NodeOfDetCharCrossroads*>(node)->matching))
            res += " shape=doublecircle";
        return res;
    }

    void append_reverse_hex(std::string& res, uint32_t num){
        if (num == 0){
            res += "0";
        } else {
            while (num){
                uint32_t r = num & 0x0F;
                res += static_cast<char>((r < 10) ? (r + '0') : (r - 10 + 'a'));
                num >>= 4;
            }
        }
    }

    std::string stringify_codeset(const codeset_t& cs){
        std::string res;
        for (long i = static_cast<long>(cs.size()) - 1; i >= 0; i--) {
            uint64_t start = cs[i].first, end = cs[i].second;
            if (start == end) {
                append_reverse_hex(res, start);
            } else {
                append_reverse_hex(res, end);
                res += '-';
                append_reverse_hex(res, start);
            }
            if (i != 0)
                res += ',';
        }
        std::reverse(res.begin(), res.end()); /* ascii works wonders */
        return res;
    }

    std::string get_extended_node_lable(FA_Node* node){
        if ((node->type == one_char_read && dynamic_cast<FA_NodeOfOneCharRead*>(node)->second_ns) ||
        (node->type == det_char_crossroads && dynamic_cast<FA_NodeOfDetCharCrossroads*>(node)->second_ns)) {
            return std::string(" ") + STAR;
        }
        if (node->type == match) {
            FA_NodeOfMatch* mn = static_cast<FA_NodeOfMatch*>(node);
            if (mn->ext_filter_added)
                return std::string(" pending loa ") + stringify_codeset(mn->pending_filter);
        }
        return "";
    }

    std::string get_node_lable(FA_Node* node, const NodesProblems& bd){
        std::string res;
        switch (node->type) {
#define tcase(tn, str) case tn: res = str; break;
            tcase(one_char_read, "ocr")
            tcase(match, "m")
            tcase(forking, "f")
            tcase(look_one_behind, "lob")
            tcase(look_one_ahead, "loa")
            tcase(track_array_mov_imm, "tami")
            tcase(track_array_mov_halfinvariant, "tamh")
            tcase(det_char_crossroads, "dcc")
        }
        res += ("[" + std::to_string(node->nodeId) + "]");
        res += get_extended_node_lable(node);
        if (bd.refcount_problem)
            res += ("!refcount: " + std::to_string(node->refs) + "!");
        return res;
    }

    void print_edge(FA_Node* start, const FA_Node* dest, const std::string& label, FILE* fd, NodesProblems& np){
        if (!dest){
            fprintf(stderr, "NULL transition going from node %lu\n", start->nodeId);
            fprintf(fd, "%lu->NULL_%lu_%lu [label=\"%s\" color=crimson]", start->nodeId,
                start->nodeId, np.edges_point_to_null++, label.c_str());
            return;
        }
        fprintf(fd, "%lu->%lu [label=\"%s\" %s]\n", start->nodeId, dest->nodeId, label.c_str(),
                get_applied_edge_attributes(start, np, EdgesProblems(false)).c_str());
    }

    void print_fa(const FA_Container& fa, FILE* fd, const KnownTrackingTools& ktr,
                const RegexPriorityTable& priority_table){
        assert(fa.start);
        assert(fd);
        fprintf(fd, "digraph finite_automaton {\ngraph ["
           "fontname = \"Helvetica\"  charset = \"UTF-8\" label = \"Finite Automaton\" labelloc = \"t\" labeljust = \"c\"  "
           "bgcolor = \"#FFFAF4\" fontcolor = black fontsize = 18 style = \"filled\" rankdir = LR margin = 0.2  "
           "splines = spline nodesep = 0.9 ranksep = 1.2 ]\n node [ style = \"solid,filled\" fontsize = 15  "
           "fontcolor = black fontname = \"Helvetica\" color = black fillcolor = white  margin = \"0.2,0.2\" shape=circle  "
           "]\n edge [ style = solid fontsize = 16 fontcolor = black fontname = \"Helvetica\" color = black  "
           "labelfloat = false labeldistance = 2.5 labelangle = 70 arrowhead = normal ]\n"
           "start_state [label = \"start\\nfrom\\nhere\" shape=none style=\"\" ]\n");

        size_t n = fa.all.size();
        std::vector<NodesProblems> breakdown;
        breakdown.resize(n);
        breakdown[fa.start->nodeId].actual_refcount++;
        for (size_t i = 0; i < n; i++){
            assert(fa.all[i]->nodeId == static_cast<int64_t>(i));
            for (FA_Node** nxtN: fa.all[i]->get_all_transitions())
                if ((*nxtN) != NULL)
                    breakdown[(**nxtN).nodeId].actual_refcount++;
        }
        for (size_t i = 0; i < n; i++){
            if (fa.all[i]->refs != breakdown[i].actual_refcount){
                breakdown[i].refcount_problem = true;
                fprintf(stderr, "Corrupted FA: wrong refcount on node %lu\n", fa.all[i]->nodeId);
            }
        }
        for (size_t i = 0; i < n; i++){
            fprintf(fd, "%lu [label=\"%s\" %s]\n", i, get_node_lable(fa.all[i], breakdown[i]).c_str(),
                    get_applied_node_attributes(fa.all[i], breakdown[i]).c_str());
        }

        /* Two Infoboxes */

        auto stringifyTrackingVarType = [](tracking_var_type_t type) -> std::string {
            switch (type) {
                case tracking_var_types::range:
                    return "range";
                case tracking_var_types::dot_cur_pos:
                    return "dot of cur pos";
                default:
                    return "dot of immediate";
            }
        };

        std::string infoText;
        for (auto& p: ktr.track_names){
            const SubtrackingNameInfo& tu = ktr.retrieval_info[p.second];

            auto getRole = [](bool presence, tracking_var_type_t type, int first, int second,
                    const std::string&  ARR_NAME) -> std::string {
                if (!presence) {
                    assert(first == -1 && second == -1);
                    return "Not involved in " + ARR_NAME;
                }
                if (type == tracking_var_types::range){
                    assert(first != -1 && second != -1);
                    return "In " + ARR_NAME + ": " + std::to_string(first) + " &lt;&minus;&gt; " + std::to_string(second);
                }
                assert(first != -1 && second == -1);
                return "In " + ARR_NAME + ": ( " + std::to_string(first) + " )";
            };
            char buf[2048] = {0};
            snprintf(buf, 2048, "Tracking unit name: %s\\n"  "Discovered: %s\\n"  "Type: %s\\n"  "%s\\n%s",
                                p.first.c_str(), tu.discovered ? "ofcourse" : "no",
                                stringifyTrackingVarType(tu.type).c_str(),
                                getRole(tu.stored_in_ca, tu.type, tu.colarr_first, tu.colarr_second, "colarr").c_str(),
                                getRole(tu.stored_in_sa, tu.type, tu.selarr_first, tu.selarr_second, "selarr").c_str());
            if (!infoText.empty())
                infoText += "|";
            infoText += buf;
        }
        fprintf(fd, "infoBoard1 [label=\"%s\" shape = record]\n", infoText.c_str());
        infoText = "";
        for (size_t i = 0; i < priority_table.size(); i++){
            const RegexPriorityTableAction& tu = priority_table[i];
            if (!infoText.empty())
                infoText += "|";
            infoText += tu.minimize ? "Minimize " : "Maximize ";
            if (tu.pos.isForRange()){
                infoText += "[" + std::to_string(tu.pos.second) + "] - [" + std::to_string(tu.pos.first) + "]";
            } else {
                infoText += "[" + std::to_string(tu.pos.first) + "]";
            }
        }
        fprintf(fd, "infoBoard2 [label=\"%s\" shape = record]\n", infoText.c_str());

        assert(fa.start);
        fprintf(fd, "start_state->%lu [color=gray style=dotted]\n", fa.start->nodeId);


        for (FA_Node* node: fa.all){
            NodesProblems& bd = breakdown[node->nodeId];
            if (node->type == one_char_read){
                FA_NodeOfOneCharRead* cn = dynamic_cast<FA_NodeOfOneCharRead *>(node);
                std::string str = stringify_codeset(cn->filter);
                print_edge(node, cn->nxt_node, str + (cn->second_ns ? std::string(" ") + STAR : ""), fd, bd);
            } else if (node->type == forking){
                FA_NodeOfForking* cn = dynamic_cast<FA_NodeOfForking *>(node);
                for (FA_Node* nxt: cn->nxt_options){
                    print_edge(node, nxt, "", fd, bd);
                }
            } else if (node->type == look_one_behind){
                FA_NodeOfLookOneBehind* cn = dynamic_cast<FA_NodeOfLookOneBehind *>(node);
                print_edge(node, cn->nxt_node, stringify_codeset(cn->filter), fd, bd);
            } else if (node->type == look_one_ahead){
                FA_NodeOfLookOneAhead* cn = dynamic_cast<FA_NodeOfLookOneAhead *>(node);
                print_edge(node, cn->nxt_node, stringify_codeset(cn->restriction), fd, bd);
            } else if (node->type == track_array_mov_imm){
                FA_NodeOfTrackArrayMovImm* cn = dynamic_cast<FA_NodeOfTrackArrayMovImm *>(node);
                char buf[1024];
                if (!isImmMovOpcode(cn->operation))
                    fprintf(stderr, "bad operation in node %lu\n", node->nodeId);
                snprintf(buf, 1024, "%s %hu %lu",
                         opcode_to_str(cn->operation), cn->key, cn->imm_value);
                print_edge(node, cn->nxt_node,std::string(buf), fd, bd);
            } else if (node->type == track_array_mov_halfinvariant){
                FA_NodeOfTrackArrayMovHalfinvariant* cn = dynamic_cast<FA_NodeOfTrackArrayMovHalfinvariant *>(node);
                char buf[1024];
                if (!isCurPosMovOpcode(cn->operation))
                    fprintf(stderr, "bad operation in node %lu\n", node->nodeId);
                snprintf(buf, 1024, "%s %hu",
                         opcode_to_str(cn->operation), cn->key);
                print_edge(node, cn->nxt_node,std::string(buf), fd, bd);
            } else if (node->type == det_char_crossroads){
                FA_NodeOfDetCharCrossroads* cn = dynamic_cast<FA_NodeOfDetCharCrossroads *>(node);
                for (const auto& transition: cn->crossroads){
                    std::string str = stringify_codeset(transition.input);
                    print_edge(node, transition.nxt_node, str + (cn->second_ns ? std::string(" ") + STAR : ""),
                               fd, bd);
                }
            }
        }
        fprintf(fd, "}\n");
    }

    FILE* get_fd(const char* apath){
        errno = 0;
        FILE *fd = fopen(apath, "w");
        if (!fd)
            perror("fopen w");
        if (ftruncate(fileno(fd), 0) != 0)
            perror("truncation");
        fd = fopen(apath, "a");
        if (!fd)
            perror("fopen a");
        return fd;
    }

    void show_fa_with_sxiv_after_dot(const FA_Container& fa, const KnownTrackingTools& ktr,
                const RegexPriorityTable& priority_table) {
        const char* temp_gv = "FAGraph.gv";
        const char* temp_png = "FAGraph.png";
        int temp_descriptor = open(temp_gv, O_CLOEXEC | O_APPEND | O_CREAT | O_WRONLY, S_IRWXU | S_IRWXG);
        assert(temp_descriptor >= 0);
        assert(fa.start);
        FILE* fd = get_fd(temp_gv);
        print_fa(fa, fd, ktr, priority_table);
        fclose(fd);
        char cmdBuf[1024];
        // todo: get rid of temporary dot file and shell usage
        snprintf(cmdBuf, 1024, "dot %s -Tpng >%s", temp_gv, temp_png);
        int chw = system(cmdBuf);
        assert(WIFEXITED(chw));
        assert(WEXITSTATUS(chw) == 0);
        snprintf(cmdBuf, 1024, "sxiv %s", temp_png);
        chw = system(cmdBuf);
        assert(WIFEXITED(chw));
        assert(WEXITSTATUS(chw) == 0);
        assert(chw >= 0);
        unlink(temp_gv);
        unlink(temp_png);
    }
}
