#ifndef LIBREGEXIS024_FINITE_AUTOMATON_H
#define LIBREGEXIS024_FINITE_AUTOMATON_H

#include <stdint.h>
#include <vector>
#include <libregexis024fa/codeset.h>
#include <libregexis024vm/vm_opcodes.h>

namespace regexis024 {
    enum FA_Node_type: uint8_t {
        match,
        one_char_read,
        forking,
        look_one_behind,
        look_one_ahead,
        track_array_mov_imm,
        track_array_mov_halfinvariant,
        /* Used for DFA */
        det_char_crossroads,
    };

    struct FA_Node{
        size_t refs = 0;
        /* If node is not in searched subset (at least yet), `search mark == -1`, otherwise
        * it is an index (for that particular node) in the vector that captures all nodes in
        * searched subset*/
        int64_t search_mark = -1;
        FA_Node_type type;
        int64_t nodeId;

        bool empty();
        virtual std::vector<FA_Node**> get_all_empty_valid_transitions();
        virtual void apply_lookahead_restriction(const codeset_t &restriction);
        void reAdd_references();
        virtual ~FA_Node() = default;
        virtual std::vector<FA_Node**> get_all_transitions();
    };

    struct FA_NodePathPart: public FA_Node{
        FA_Node* nxt_node = NULL;

        std::vector<FA_Node **> get_all_empty_valid_transitions() override;
        std::vector<FA_Node **> get_all_transitions() override;
    };

    struct FA_NodeOfMatch: public FA_Node{
        bool ext_filter_added = false;
        codeset_t pending_filter;

        explicit FA_NodeOfMatch();
        void apply_lookahead_restriction(const codeset_t &restriction) override;
    };

    /* .type == one_char_read */
    struct FA_NodeOfOneCharRead: public FA_NodePathPart{
        codeset_t filter;
        bool second_ns = false;

        FA_NodeOfOneCharRead(const codeset_t &filter, bool second_namespace);
        void apply_lookahead_restriction(const codeset_t &restriction) override;
        std::vector<FA_Node **> get_all_empty_valid_transitions() override;
    };

    /* .type == forking */
    struct FA_NodeOfForking: public FA_Node{
        /* Won't be modified after init (in regexp compilation into NFA) */
        std::vector<FA_Node*> nxt_options;
        int64_t stopId = -1;

        explicit FA_NodeOfForking();
        std::vector<FA_Node **> get_all_empty_valid_transitions() override;
        std::vector<FA_Node **> get_all_transitions() override;
    };

    /* .type == look_one_behind */
    struct FA_NodeOfLookOneBehind: public FA_NodePathPart{
        /* [0; UINT32_MAX] is equivalent to no filter */
        codeset_t filter;

        explicit FA_NodeOfLookOneBehind(const codeset_t &filter);
    };

    /* .type == look_one_ahead */
    struct FA_NodeOfLookOneAhead: public FA_NodePathPart{
        /* [0; UINT32_MAX] is equivalent to no restriction */
        codeset_t restriction;

        explicit FA_NodeOfLookOneAhead(const codeset_t &restriction);
    };

    /* .type == track_array_mov_imm */
    struct FA_NodeOfTrackArrayMovImm: public FA_NodePathPart{
        opcode_t operation;
        uint16_t key;
        uint64_t imm_value;

        FA_NodeOfTrackArrayMovImm(opcode_t operation, uint16_t key, uint64_t immValue);
    };

    /* .type == track_array_mov_halfinvariant */
    struct FA_NodeOfTrackArrayMovHalfinvariant: public FA_NodePathPart{
        opcode_t operation;
        uint16_t key;

        FA_NodeOfTrackArrayMovHalfinvariant(opcode_t operation, uint16_t key);
    };

    struct DFA_CrossroadPath{
        codeset_t input;
        FA_Node* nxt_node = NULL;

        DFA_CrossroadPath(const codeset_t &input, FA_Node *nxt_node);
        DFA_CrossroadPath() = default;
    };

    /* .type == det_char_crossroads */
    struct FA_NodeOfDetCharCrossroads: public FA_Node{
        std::vector<DFA_CrossroadPath> crossroads;
        bool matching = false;
        bool second_ns = false;

        explicit FA_NodeOfDetCharCrossroads(const std::vector<DFA_CrossroadPath> &crossroads);
        void apply_lookahead_restriction(const codeset_t &restriction) override;
        std::vector<FA_Node **> get_all_empty_valid_transitions() override;
        std::vector<FA_Node **> get_all_transitions() override;
    };

    struct FA_Container{
        FA_Container(const FA_Container&) = delete;
        FA_Container& operator=(const FA_Container&) = delete;
        FA_Container() = default;

        std::vector<FA_Node*> all;
        FA_Node* start = NULL;

        void registerNew(FA_Node* node);

        FA_NodeOfMatch* makeMatch();
        FA_NodeOfOneCharRead* makeOneCharRead(const codeset_t& filter, bool second_namespace);
        FA_NodeOfForking* makeForking();
        FA_NodeOfLookOneBehind* makeLookOneBehind(const codeset_t& filter);
        FA_NodeOfLookOneAhead* makeLookOneAhead(const codeset_t& filter);
        FA_NodeOfTrackArrayMovImm* makeTrackArrayMovImm(opcode_t operation, uint16_t key, uint64_t immValue);
        FA_NodeOfTrackArrayMovHalfinvariant* makeTrackArrayMovHalfinvariant(opcode_t operation, uint16_t key);
        FA_NodeOfDetCharCrossroads* makeDetCharCrossroads();

        ~FA_Container();
    };
}

#endif //LIBREGEXIS024_FINITE_AUTOMATON_H
