#include <libregexis024sol/square_bracket_expression.h>
#include <libregexis024sol/sol_misc_base.h>
#include <libregexis024sol/special_terminals.h>

#include <vector>
#include <memory>
#include <assert.h>

namespace regexis024 {
    /* Can allow backslash (later should check that backslash expression is not multicharar or empty */
    bool soundsLikeCharOrRangeStart(int32_t peeped) {
        return peeped >= 0 && (peeped != U'[' && peeped != U']' && peeped != U'!' && \
        peeped != '^' && peeped != '&' && peeped != '-');
    }

    typedef REGEX_IS024_MeaningContext ctx_t;

    struct ParseCall;
    typedef std::shared_ptr<ParseCall> chekushka;

    struct ParseCall{
        codeset_t& result;

        explicit ParseCall(codeset_t &result) : result(result) {}
        virtual ~ParseCall() = default;
        virtual chekushka afterReceive(ctx_t& ctx, const CommonCodesets& cc) { assert(false); }
        virtual chekushka firstTime(ctx_t& ctx, const CommonCodesets& cc) { assert(false); }
    };

#define call_ERROR_CHECK do { if (ctx.error) { return NULL; } } while (0)
#define call_THROW(str) do { report(ctx, "square bracket expression: " str); return NULL; } while (0)

    /* [...] */
    struct ZeroLvl_ParseCall: public ParseCall{
        explicit ZeroLvl_ParseCall(codeset_t &result) : ParseCall(result) {}
        chekushka afterReceive(ctx_t &ctx, const CommonCodesets& cc) override;
        chekushka firstTime(ctx_t &ctx, const CommonCodesets& cc) override;
    };

    /* ...&...&... */
    struct FirstLvl_ParseCall: public ParseCall{
        codeset_t ret_buf_for_new;
        bool got_one = false;
        explicit FirstLvl_ParseCall(codeset_t& result) : ParseCall(result) {}
        chekushka afterReceive(ctx_t &ctx, const CommonCodesets& cc) override;
        chekushka firstTime(ctx_t &ctx, const CommonCodesets& cc) override;
    };

    /* ab[]vgd[]eyo[]zhz */
    struct SecondLvl_ParseCall: public ParseCall{
        codeset_t ret_buf_for_new;
        explicit SecondLvl_ParseCall(codeset_t& result) : ParseCall(result) {}
        chekushka afterReceive(ctx_t &ctx, const CommonCodesets& cc) override;
        chekushka firstTime(ctx_t &ctx, const CommonCodesets& cc) override;
    };

    /* ^... */
    struct CircumflexLvl_ParseCall: public ParseCall{
        codeset_t ret_buf_for_new;
        explicit CircumflexLvl_ParseCall(codeset_t& result) : ParseCall(result) {}
        chekushka afterReceive(ctx_t &ctx, const CommonCodesets& cc) override;
        chekushka firstTime(ctx_t &ctx, const CommonCodesets& cc) override;
    };

    /* ********* ZeroLvl_ParseCall ********** */

    chekushka ZeroLvl_ParseCall::firstTime(ctx_t &ctx, const CommonCodesets& cc) {
        assert(readChar(ctx) == U'[');
        return std::make_shared<FirstLvl_ParseCall>(result);
    }

    chekushka ZeroLvl_ParseCall::afterReceive(ctx_t &ctx, const CommonCodesets& cc) {
        if (peep(ctx) != U']')
            call_THROW("lvl 0: missing ]");
        readChar(ctx);
        return NULL;
    }

    /* ********* FirstLvl_ParseCall ********** */

    chekushka FirstLvl_ParseCall::firstTime(ctx_t &ctx, const CommonCodesets& cc) {
        return std::make_shared<SecondLvl_ParseCall>(result);
    }

    chekushka FirstLvl_ParseCall::afterReceive(ctx_t &ctx, const CommonCodesets& cc) {
        if (got_one)
            result = intersect_sets(result, ret_buf_for_new);
        else
            got_one = true;
        if (peep(ctx) == U'&'){
            readChar(ctx);
            return std::make_shared<SecondLvl_ParseCall>(ret_buf_for_new);
        }
        return NULL;
    }

    /* ********* SecondLvl_ParseCall ********** */

    chekushka SecondLvl_ParseCall::firstTime(ctx_t &ctx, const CommonCodesets& cc) {
        repeat:
        int32_t ch = peep(ctx); call_ERROR_CHECK;
        if (ch == U'^'){
            return std::make_shared<CircumflexLvl_ParseCall>(ret_buf_for_new);
        } else if (ch == U'!'){
            Command cmd = command_expr_parse(ctx); call_ERROR_CHECK;
            if (!is_command_for_charset(cmd))
                call_THROW("second lvl: illegal command");
            interpret_command_as_charset_giving(cc, cmd, ret_buf_for_new);
            result = merge_sets(result, ret_buf_for_new);
            goto repeat;
        } else if (ch == U'['){
            return std::make_shared<ZeroLvl_ParseCall>(ret_buf_for_new);
        } else if (soundsLikeCharOrRangeStart(ch)){
            readChar(ctx);
            bool bs_multicode;
            codeset_t bs_stuff;

            if (ch == '\\'){
                backslash_expression_parsing_try_regular(ctx, cc, bs_multicode, bs_stuff);
                if (bs_multicode){
                    result = merge_sets(result, bs_stuff);
                    goto repeat;
                } else {
                    ret_buf_for_new = codeset_of_one_char(bs_stuff[0].first);
                }
            } else {
                ret_buf_for_new = codeset_of_one_char(ch);
            }
            int32_t mCh = peep(ctx); call_ERROR_CHECK;
            if (mCh == U'-'){
                readChar(ctx);
                int32_t scnd = peep(ctx); call_ERROR_CHECK;
                readChar(ctx);
                if (scnd == U'\\'){
                    backslash_expression_parsing_try_regular(ctx, cc, bs_multicode, bs_stuff);
                    if (bs_multicode)
                        call_THROW("second lvl: char range: bad escape expression after hyphen");
                    ret_buf_for_new[0].second = bs_stuff[0].first;
                } else if (soundsLikeCharOrRangeStart(scnd)){
                    ret_buf_for_new[0].second = (uint32_t)scnd;
                } else {
                    call_THROW("second lvl: char range: bad value after hyphen");
                }
                if (ret_buf_for_new[0].second < ret_buf_for_new[0].first)
                    call_THROW("second: lvl: char range: invalid range");
            }
            result = merge_sets(result, ret_buf_for_new);
            goto repeat;
        }
        return NULL;
    }

    chekushka SecondLvl_ParseCall::afterReceive(ctx_t &ctx, const CommonCodesets& cc) {
        result = merge_sets(result, ret_buf_for_new);
        return firstTime(ctx, cc);
    }

    /* ********* CircumflexLvl_ParseCall ********* */

    chekushka CircumflexLvl_ParseCall::firstTime(ctx_t &ctx, const CommonCodesets& cc) {
        assert(readChar(ctx) == U'^');
        return std::make_shared<FirstLvl_ParseCall>(ret_buf_for_new);
    }

    chekushka CircumflexLvl_ParseCall::afterReceive(ctx_t &ctx, const CommonCodesets& cc) {
        result = invert_set(ret_buf_for_new);
        return NULL;
    }


    /* Aaaaaaaaand... The function we have all been waiting for so long! */
    codeset_t sq_bracket_expr_parse(REGEX_IS024_MeaningContext &ctx, const CommonCodesets& cc) {
        std::vector<std::shared_ptr<ParseCall>> callStack;
        codeset_t res;
        callStack.push_back(std::make_shared<ZeroLvl_ParseCall>(res));
        bool first_time = true;
        while (!callStack.empty()){
            if (ctx.error)
                return {};
            auto nxt = first_time ? callStack.back()->firstTime(ctx, cc) : callStack.back()->afterReceive(ctx, cc);
            if (nxt){
                callStack.push_back(nxt);
                first_time = true;
            } else {
                callStack.pop_back();
                first_time = false;
            }
        }
        return res;
    }
}
