#include <libregexis024sol/subexpr_fa_transformed.h>
#include <libregexis024fa/misc_fa_funcs.h>
#include <assert.h>
#include <stdio.h>

namespace regexis024 {
    SubExprCompiled subexpr_charset_reading_filter(const codeset_t &codeset, FA_Container &fa) {
        return subexpression_from_path(fa.makeOneCharRead(codeset, false));
    }

    SubExprCompiled join(const SubExprCompiled &A, const SubExprCompiled &B) {
        if (!A.start)
            return B;
        if (!B.start)
            return A;
        SubExprCompiled res;
        res.start = A.start;
        for (FA_Node** ptrToptr : A.ends)
            reattach_fa_node_edge(ptrToptr, B.start);
        res.ends = B.ends;
        res.can_be_empty = A.can_be_empty && B.can_be_empty;
        return res;
    }

    SubExprCompiled subexpression_from_path(FA_NodePathPart *node) {
        SubExprCompiled res;
        res.start = node;
        res.ends.push_back(&(node->nxt_node));
        /* There is only one char reading path node type */
        res.can_be_empty = (node->type != one_char_read);
        return res;
    }

    SubExprCompiled RobertAngier(const SubExprCompiled& source, FA_Container& fa) {
        SubExprCompiled res;
        if (!source.start)
            return res;

        struct Marked{
            FA_Node *original = NULL, *clone = NULL;
            explicit Marked(FA_Node *original) : original(original) {}
        };
        std::vector<Marked> searched;
        searched.push_back(Marked(source.start));
        source.start->search_mark = 0;

        for (size_t done = 0; done < searched.size(); done++){
            FA_Node& v = *searched[done].original;
            searched[done].clone = copy_fa_node(v, fa);
            for (FA_Node **nxtN: searched[done].clone->get_all_transitions()){
                if (!(*nxtN))
                    res.ends.push_back(nxtN);
                else if ((**nxtN).search_mark < 0){
                    (**nxtN).search_mark = (int64_t)searched.size();
                    searched.emplace_back(*nxtN);
                }
            }
        }
        res.start = searched[0].clone;
        for (Marked& mrkd: searched){
            for (FA_Node **nxtN: mrkd.clone->get_all_transitions()){
                if (*nxtN){
                    assert((**nxtN).search_mark >= 0);
                    Marked& proc_nxt = searched[(**nxtN).search_mark];
                    reattach_fa_node_edge(nxtN, proc_nxt.clone);
                }
            }
        }
        for (Marked& mrkd: searched)
            mrkd.original->search_mark = -1;
        return res;
    }

    void reattach_all_ends_to_one_node(SubExprCompiled& patient, FA_Node* node){
        assert(node);
        assert(patient.start);
        for (FA_Node** end: patient.ends){
            assert(!(*end));
            printf("DEBUG %lu->->->->->%lu\n", patient.start->nodeId, node->nodeId);
            reattach_fa_node_edge(end, node);
        }
    }

    void apply_repeat_to_subexpression(SubExprCompiled &patient, FA_Container& fa, size_t min_allowed, size_t max_allowed) {
        assert(min_allowed <= max_allowed && min_allowed <= REGEXIS024_MAX_REPEAT);
        if (!patient.start)
            return;
        bool infinite_repeat = max_allowed > REGEXIS024_MAX_REPEAT;
        if (min_allowed == 0 && max_allowed == 0){
            patient = {};
        } else if (min_allowed == 1 && max_allowed == 1){
            /* Chill */
        } else if (min_allowed == 0 && infinite_repeat){
            FA_NodeOfForking* fn = fa.makeForking();
            add_option_to_fork_node(fn, patient.start);
            for (FA_Node** old_end: patient.ends)
                reattach_fa_node_edge(old_end, fn);
            add_option_to_fork_node(fn, NULL);
            patient.start = fn;
            patient.ends = {&(fn->nxt_options[1])};
        } else if (min_allowed == 1 && infinite_repeat) {
            FA_NodeOfForking* fn = fa.makeForking();
            reattach_all_ends_to_one_node(patient, fn);
            add_option_to_fork_node(fn, patient.start);
            add_option_to_fork_node(fn, NULL);
            patient.ends = {&(fn->nxt_options[1])};
        } else if (min_allowed == 0 && max_allowed == 1){
            FA_NodeOfForking* fn = fa.makeForking();
            add_option_to_fork_node(fn, patient.start);
            add_option_to_fork_node(fn, NULL);
            patient.start = fn;
            patient.ends.push_back(&(fn->nxt_options[1]));
        } else if (infinite_repeat) {
            std::vector<SubExprCompiled> Colon(min_allowed);
            Colon[0] = patient;
            for (size_t i = 1; i < min_allowed; i++)
                Colon[i] = RobertAngier(patient, fa);
            FA_NodeOfForking* fn = fa.makeForking();
            for (size_t i = 0; i + 1 < min_allowed; i++)
                reattach_all_ends_to_one_node(Colon[i], Colon[i + 1].start);
            reattach_all_ends_to_one_node(Colon[min_allowed - 1], fn);
            add_option_to_fork_node(fn, Colon[min_allowed - 1].start);
            add_option_to_fork_node(fn, NULL);
            /* patient.start is the same (the original is at Colon[0] */
            patient.ends = {&(fn->nxt_options[1])};
        } else {
            std::vector<SubExprCompiled> Avenue(max_allowed);
            Avenue[max_allowed - 1] = patient;
            for (size_t i = 0; i < max_allowed - 1; i++)
                Avenue[i] = RobertAngier(patient, fa);
            for (size_t i = 0; i + 1 < max_allowed; i++)
                reattach_all_ends_to_one_node(Avenue[i], Avenue[i + 1].start);
            FA_NodeOfForking* fn = fa.makeForking();
            if (min_allowed > 0){
                for (size_t i = 0; i <= max_allowed - min_allowed; i++)
                    add_option_to_fork_node(fn, Avenue[i].start);
            } else {
                for (size_t i = 0; i < max_allowed; i++)
                    add_option_to_fork_node(fn, Avenue[i].start);
                add_option_to_fork_node(fn, NULL);
                patient.ends.push_back(&(fn->nxt_options[max_allowed]));
            }
            patient.start = fn;
            /* patient.ends is the same (the original is Avenue.back()) */
        }
        if (min_allowed == 0)
            patient.can_be_empty = true;
    }

    SubExprCompiled forkify(const std::vector<SubExprCompiled> &options, FA_Container& fa){
        SubExprCompiled result;
        size_t non_empty = 0;
        result.can_be_empty = false;
        for (const SubExprCompiled& opt: options){
            result.can_be_empty |= opt.can_be_empty;
            if (opt.start)
                non_empty++;
        }
        if (non_empty == 0){
            result.can_be_empty = true;
            return result;
        }
        if (non_empty == 1){
            for (const SubExprCompiled& opt: options)
                if (opt.start){
                    result = opt;
                    break;
                }
        } else  {
            FA_NodeOfForking* n1 = fa.makeForking();
            result.start = n1;
            n1->nxt_options.reserve(non_empty);
            for (const SubExprCompiled& opt: options)
                if (opt.start){
                    add_option_to_fork_node(n1, opt.start);
                    for (FA_Node** end: opt.ends)
                        result.ends.push_back(end);
                }
        }
        return result;
    }

    void SubExprCompiled::assertDefault() {
        assert(!start && ends.empty() && can_be_empty);
    }
}
