#include <libregexis024fa/colored_codeset.h>
#include <assert.h>
#include <string>
#include <algorithm>
#include <map>
#include <exception>
#include <stdexcept>
#include <random>

using namespace regexis024;

struct test_id_t {
    int test_id;
    int subtest_id;

    test_id_t(int test_id, int subtest_id) : test_id(test_id),subtest_id(subtest_id) {}

    std::string toString() const {
        char buf[128];
        snprintf(buf, 128, "#%d::%d", test_id + 1, subtest_id + 1);
        return buf;
    }
};

std::string stringifyCodeset(const codeset_t& CS) {
    std::string cs;
    for (auto p: CS) {
        if (!cs.empty())
            cs += "; ";
        cs += std::to_string(p.first) + "-" + std::to_string(p.second);
    }
    return cs;
}

std::string stringifyRequestList(const std::vector<uint64_t>& arr) {
    std::string rl;
    for (auto r: arr) {
        if (!rl.empty())
            rl += ", ";
        rl += std::to_string(r);
    }
    return rl;
}

void print_obj(const std::vector<std::pair<codeset_t, std::vector<size_t>>>& answer) {
    size_t R = answer.size();
    for (int i = 0; i < R; i++) {
        std::string cs = stringifyCodeset(answer[i].first);
        std::string rl = stringifyRequestList(answer[i].second);
        printf("{%s} -> {%s}\n", cs.c_str(), rl.c_str());
    }
}

void print_answer_canonic(const std::map<std::vector<size_t>, codeset_t>& answer) {
    for (auto& p: answer) {
        printf("{%s} -> {%s}\n", stringifyCodeset(p.second).c_str(), stringifyRequestList(p.first).c_str());
    }
}

void print_test_details(int dummyN, const std::vector<codeset_t>& requests) {
    for (int i = 0; i < requests.size(); i++) {
        printf("%d) %s\n", i - dummyN, stringifyCodeset(requests[i]).c_str());
    }
}

void fail_test(test_id_t my_test_id) {
    char buf[1024];
    snprintf(buf, 1024, "Test %s failed", my_test_id.toString().c_str());
    throw std::runtime_error(buf);
}

std::map<std::vector<size_t>, codeset_t> safe_canonificate_answer(test_id_t my_test_id,
    const std::vector<std::pair<codeset_t, std::vector<size_t>>>& answer)
{
    std::map<std::vector<size_t>, codeset_t> answer_canonic;
    for (auto& p: answer) {
        if (answer_canonic.count(p.second) == 0)
            answer_canonic[p.second] = {};
        if (!intersect_sets(answer_canonic[p.second], p.first).empty())
            fail_test(my_test_id);
        answer_canonic[p.second] = merge_sets(answer_canonic[p.second], p.first);
    }
    return answer_canonic;
}

std::vector<std::pair<codeset_t, std::vector<size_t>>> safe_zip_answer(test_id_t my_test_id,
    const std::vector<codeset_t>& ti, const std::vector<std::vector<size_t>>& to)
{
    size_t R = ti.size();
    std::vector<std::pair<codeset_t, std::vector<size_t>>> res(R);
    if (to.size() != R)
        fail_test(my_test_id);
    for (size_t i = 0; i < R; i++) {
        res[i].first = ti[i];
        res[i].second = to[i];
    }
    return res;
}

void perform_test(test_id_t my_test_id, int InpDummyN, const std::vector<codeset_t>& rqInpCs, const std::vector<std::pair<codeset_t, std::vector<size_t>>>& answer_right)
{
    std::map<std::vector<size_t>, codeset_t> answer_canonic_right = safe_canonificate_answer(my_test_id, answer_right);
    ColoredCodeset cc(InpDummyN);

    for (auto& c: rqInpCs)
        cc.apply_divisor(c);
    std::vector<codeset_t> ti;
    std::vector<std::vector<size_t>> to;
    cc.get_splits_of_non_dummy(ti, to);
    std::vector<std::pair<codeset_t, std::vector<size_t>>> answer_returned = safe_zip_answer(my_test_id, ti, to);
    std::map<std::vector<size_t>, codeset_t> answer_canonic_returned = safe_canonificate_answer(my_test_id,
        answer_returned);
    if (answer_canonic_right != answer_canonic_returned) {
        printf("Test failed!!\n");
        printf("Test details:\n");
        print_test_details(InpDummyN, rqInpCs);
        printf("Right answer:\n");
        print_obj(answer_right);
        printf("Given answer:\n");
        print_obj(answer_returned);
        fail_test(my_test_id);
    }
    printf("Test %s passed\n", my_test_id.toString().c_str());
}

void perform_test_with_shuffle(const std::vector<codeset_t>& rqInpCs,
    const std::vector<std::pair<codeset_t, std::vector<size_t>>>& answer)
{
    static int cur_test_id = 0;
    int my_test_id = cur_test_id++;
    int dn = rqInpCs.size();
    std::vector<int> shuf(dn);
    for (int i = 0; i < dn; i++)
        shuf[i] = i;
    for (int stid = 0; stid < 14; stid++) {
        std::random_device d;
        std::mt19937 r_gen(d());
        std::shuffle(shuf.begin(), shuf.end(), r_gen);
        int InpDummyN = std::uniform_int_distribution<int>(0, dn - 1)(r_gen);
        // for (int e: shuf)
            // printf("%d ", e);
        // printf("\n");
        std::vector<codeset_t> new_rqInpCs(dn);
        for (int i = 0; i < dn; i++)
            new_rqInpCs[shuf[i]] = rqInpCs[i];
        std::vector<std::pair<codeset_t, std::vector<size_t>>> new_right_answer;
        for (const std::pair<codeset_t, std::vector<size_t>>& p: answer) {
            std::vector<uint64_t> new_request_list;
            for (size_t oldRid: p.second) {
                size_t unabridged_new_rid = shuf[oldRid];
                if (unabridged_new_rid >= InpDummyN) {
                    new_request_list.push_back(unabridged_new_rid - InpDummyN);
                }
            }
            if (!new_request_list.empty()) {
                std::sort(new_request_list.begin(), new_request_list.end());
                new_right_answer.emplace_back(p.first, new_request_list);
            }
        }
        perform_test(test_id_t(my_test_id, stid), InpDummyN, new_rqInpCs, new_right_answer);
    }
}

int main() {
    perform_test_with_shuffle(
        {
            {{5, 500}},
            {{10, 20}},
            {{5, 10}},
        },
        {
            {{{21, 500}}, {0}},
            {{{11, 20}}, {0, 1}},
            {{{5, 9}}, {0, 2}},
            {{{10, 10}}, {0, 1, 2}},
        });

    perform_test_with_shuffle({
        {{10, 19}},
        {{10, 15}},
        {{5, 9}},
        {{20, 40}},
        {{16, UINT32_MAX}},
    },
    {
        {{{5, 9}}, {2}},
        {{{10, 15}}, {0, 1}},
        {{{16, 19}}, {0, 4}},
        {{{20, 40}}, {3, 4}},
        {{{41, UINT32_MAX}}, {4}}

    });

    perform_test_with_shuffle(
        {
            {{10, 19}, {30, 39}, {50, 69}},
            {{20, 29}, {40, 59}, },
            {{20, 39}, {70, 79}},
            codeset_of_one_char(UINT32_MAX - 1),
            codeset_of_one_char(UINT32_MAX),
            {{UINT32_MAX - 1, UINT32_MAX}},
            codeset_of_one_char(0),
            codeset_of_one_char(1),
            {{0, 1}},
        },
        {
            {{{10, 19}, {60, 69}}, {0}},
            {{{40, 49}}, {1}},
            {{{70, 79}}, {2}},
            {{{50, 59}}, {0, 1}},
            {{{20, 29}}, {1, 2}},
            {{{30, 39}}, {0, 2}},
            {{{0, 0}}, {6, 8}},
            {{{1, 1}}, {7, 8}},
            {{{UINT32_MAX - 1, UINT32_MAX - 1}}, {3, 5}},
            {{{UINT32_MAX, UINT32_MAX}}, {4, 5}},
        });
    return 0;
}
