#include <libregexis024vm/instruction_implementation.h>
#include <stdexcept>

namespace regexis024 {
    void swap_old_settled_and_new_active(VMContext &ctx, Thread& old_settled){
        ctx_print_debug(ctx);
        assert(old_settled.slot_occupation_status == SLOT_OCCUPIED_val);
        Thread temp = old_settled;
        old_settled = ctx.active_thread;
        old_settled.slot_occupation_status = SLOT_NEW_val;
        ctx.active_thread = temp;
        // slot_occupation_status & SLOT_OCCUPIED of active thread is true, because it was retrieved from old_settled
    }

    void start_noncloning_conflict(VMContext& ctx, Thread& other){
        ctx_print_debug(ctx);
        if (ctx.have_sift_function){
            ctx.sifting_with = &other;
            ctx.who_started_sift = opcode_t::READ;
            ctx.intruder_IP = ctx.active_thread.IP;
            ctx.active_thread.IP = ctx.sift_function;
            ctx.RAX = ctx.RBX = 0;
        } else {
            ctx.active_thread.delete_thread();
            ctx.try_to_continue_scheduled();
        }
    }

    /* The one that drops as an intruder here is current active.thread.IP */
    void start_cloning_conflict(VMContext& ctx, Thread& other, near_ptr_t clone_IP){
        ctx_print_debug(ctx);
        if (ctx.have_sift_function){
            ctx.sifting_with = &other;
            ctx.who_started_sift = opcode_t::FORK;
            ctx.intruder_IP = ctx.active_thread.IP;
            ctx.child_ret_IP = clone_IP;
            ctx.active_thread.IP = ctx.sift_function;
            ctx.RAX = ctx.RBX = 0;
        } else {
            ctx.active_thread.IP = clone_IP;
        }
    }

#define initialization_phase_check()  if (ctx.initialized){ \
ctx.error = error_codes::too_late; return; }
#define general_matching_mode_check() if (!ctx.initialized){ \
ctx.error = error_codes::too_early; return; } if(ctx.sifting_with){ \
ctx.error = error_codes::instruction_not_for_collision_thread; return; }
#define sift_mode_check()   if (!ctx.sifting_with){ \
ctx.error = error_codes::instruction_not_for_collision_thread; return; }

    /* Can append to both read_halted+new stacks of context */
    void read_halted_new_type_stacks_append(VMContext &ctx, sslot_id_t ssid){
        ctx_print_debug(ctx);
        if (ssid < ctx.portion_of_FIRST_read_halt_ns){
            ctx.READ_halted_stack_new_first.append(ssid);
        } else {
            ctx.READ_halted_stack_new_second.append(ssid);
        }
    }

    void do_i_read(VMContext &ctx, sslot_id_t ssid) {
        ctx_print_debug(ctx);
        general_matching_mode_check()
        if (ssid >= ctx.read_slots_number)
            smitsya(read_sslot_out_of_range);
        Thread& other = ctx.READ_halted_slots[ssid];
        if (other.slot_occupation_status & SLOT_OCCUPIED){
            if (other.slot_occupation_status & SLOT_NEW){
                start_noncloning_conflict(ctx, other);
            } else {
                swap_old_settled_and_new_active(ctx, other);
                /* Even though ssid was registed in stack for elders, now young stack should also track this slot */
                read_halted_new_type_stacks_append(ctx, ssid);
            }
        } else {
            other = ctx.active_thread;
            other.slot_occupation_status = SLOT_NEW_val;
            ctx.active_thread.slot_occupation_status = SLOT_EMPTY_val;
            read_halted_new_type_stacks_append(ctx, ssid);
            ctx.try_to_continue_scheduled();
        }
    }

    void i_READ(VMContext &ctx) {
        ctx_print_debug(ctx);
        check_available_prg(BYTECODE_SSLOT_ID_SZ)
        sslot_id_t ssid = ctx.extract_sslot_id();
        do_i_read(ctx, ssid);
    }

    void i_READZ(VMContext &ctx) {
        ctx_print_debug(ctx);
        do_i_read(ctx, 0);
    }

    void i_JUMP(VMContext& ctx){
        ctx_print_debug(ctx);
        check_available_prg(BYTECODE_NEAR_POINTER_SZ)
        ctx.active_thread.IP = ctx.extract_near_pointer();
    }

    template<typename conditionT, typename immArgSzT>
    void i_JC(VMContext& ctx)
    {
        ctx_print_debug(ctx);
        check_available_prg(immArgSzT::byte_sz + BYTECODE_NEAR_POINTER_SZ);
        uint64_t imm_val_B = immArgSzT::extract(ctx);
        near_ptr_t dest = ctx.extract_near_pointer();
        uint64_t imm_val_A = ctx.INP;
        if (conditionT::call(imm_val_A, imm_val_B))
            ctx.active_thread.IP = dest;
    }

    struct condEqual{static bool call(uint64_t A, uint64_t B){return A == B;}};
    struct condLess{static bool call(uint64_t A, uint64_t B){return A < B;}};
    struct condGrtr{static bool call(uint64_t A, uint64_t B){return A > B;}};

    struct immArgByte{
        static constexpr int byte_sz = 1;
        static uint64_t extract(VMContext& ctx){return ctx.extract_b();}
    };
    struct immArgWord{
        static constexpr int byte_sz = 2;
        static uint64_t extract(VMContext& ctx){return ctx.extract_w();}
    };
    struct immArgDoubleWord{
        static constexpr int byte_sz = 4;
        static uint64_t extract(VMContext& ctx){return ctx.extract_dw();}
    };
    struct immArgQuadWord{
        static constexpr int byte_sz = 8;
        static uint64_t extract(VMContext& ctx){return ctx.extract_qw();}
    };

    void clone_thread_into_slot(Thread& source, Thread& vessel){
        thread_print_debug(source);
        my_assert(!(vessel.slot_occupation_status & SLOT_OCCUPIED));
        my_assert((source.slot_occupation_status & SLOT_OCCUPIED));
        vessel = source;
        if (vessel.CAHptr){
            vessel.CAHptr->refs++;
        }
        if (vessel.SAptr){
            vessel.SAptr[0]++;
        }
    }

    /* One FORK-slot governs the one single unique position in program: the next one after the fork */
    void i_FORK(VMContext& ctx){
        ctx_print_debug(ctx);
        general_matching_mode_check()
        check_available_prg(BYTECODE_SSLOT_ID_SZ + BYTECODE_NEAR_POINTER_SZ);
        sslot_id_t ssid = ctx.extract_sslot_id();
        near_ptr_t dest = ctx.extract_near_pointer();
        if (ssid >= ctx.fork_slots_number)
            smitsya(fork_sslot_out_of_range);
        Thread& other = ctx.FORK_halted_slots[ssid];
        if (other.slot_occupation_status & SLOT_OCCUPIED){
            start_cloning_conflict(ctx, other, dest);
        } else {
            clone_thread_into_slot(ctx.active_thread, other);
            ctx.active_thread.IP = dest;
            ctx.FORK_halted_stack.append(ssid);
        }
    }

    void i_MATCH(VMContext& ctx){
        ctx_print_debug(ctx);
        general_matching_mode_check()
        if (ctx.matched_thread.slot_occupation_status & SLOT_OCCUPIED){
            start_cloning_conflict(ctx, ctx.matched_thread, ctx.active_thread.IP);
        } else {
            clone_thread_into_slot(ctx.active_thread, ctx.matched_thread);
        }
    }

    void i_DIE(VMContext& ctx){
        ctx_print_debug(ctx);
        general_matching_mode_check()
        ctx.active_thread.delete_thread();
        ctx.try_to_continue_scheduled();
    }

    void i_PARAM_READ_SS_NUMBER(VMContext& ctx){
        ctx_print_debug(ctx);
        initialization_phase_check()
        check_available_prg(BYTECODE_SSLOT_ID_SZ)
        sslot_id_t read_slots_number = ctx.extract_sslot_id();
        ctx.read_slots_number = read_slots_number;
    }

    void i_PARAM_FORK_SS_NUMBER(VMContext& ctx){
        ctx_print_debug(ctx);
        initialization_phase_check()
        check_available_prg(BYTECODE_SSLOT_ID_SZ)
        sslot_id_t fork_slots_number = ctx.extract_sslot_id();
        ctx.fork_slots_number = fork_slots_number;
    }

    void i_PARAM_SELARR_LEN(VMContext& ctx){
        ctx_print_debug(ctx);
        initialization_phase_check()
        check_available_prg(BYTECODE_TRACK_ARRAY_INDEX_ID_SZ)
        tai_t selection_array_len = ctx.extract_track_array_index();
        ctx.selection_array_len = selection_array_len;
    }

    void i_PARAM_COLSIFTFUNC_SET(VMContext& ctx){
        ctx_print_debug(ctx);
        initialization_phase_check()
        check_available_prg(BYTECODE_NEAR_POINTER_SZ)
        near_ptr_t sift_function = ctx.extract_near_pointer();
        ctx.have_sift_function = true;
        ctx.sift_function = sift_function;
    }

    void i_PARAM_COLSIFTFUNC_WIPE(VMContext& ctx){
        ctx_print_debug(ctx);
        initialization_phase_check()
        ctx.have_sift_function = false;
    }

    void i_MSG_MULTISTART_ALLOWED(VMContext& ctx){
        ctx_print_debug(ctx);
        initialization_phase_check()
        check_available_prg(1)
        ctx.allows_multistart = (bool)ctx.extract_b();
    }

    void i_MSG_FED_INPUT_EXTENDED(VMContext& ctx){
        ctx_print_debug(ctx);
        initialization_phase_check()
        check_available_prg(1 + 1 + BYTECODE_SSLOT_ID_SZ)
        ctx.fed_input_extends_left = ctx.extract_b();
        ctx.fed_input_extends_right = ctx.extract_b();
        ctx.portion_of_second_read_halt_ns = ctx.extract_sslot_id();
    }

    uint64_t get_el_from_selarr(uint64_t* sa, near_ptr_t ind){
        return sa ? sa[1UL + ind] : 0;
    }

    void i_DMOV_RABX_SELARR(VMContext& ctx){
        ctx_print_debug(ctx);
        sift_mode_check()
        check_available_prg(BYTECODE_TRACK_ARRAY_INDEX_ID_SZ)
        tai_t i1 = ctx.extract_track_array_index();
        if (i1 >= ctx.selection_array_len)
            smitsya(selection_arr_out_of_range);
        ctx.RAX = get_el_from_selarr(ctx.active_thread.SAptr, i1);
        ctx.RBX = get_el_from_selarr(ctx.sifting_with->SAptr, i1);
    }

    uint64_t get_selarr_el_dist(uint64_t* sa, uint16_t start, uint16_t end){
        uint64_t v_start = get_el_from_selarr(sa, start);
        uint64_t v_end = get_el_from_selarr(sa, end);
        return v_end > v_start ? v_end - v_start : 0;
    }

    void i_DDIST_RABX_SELARR(VMContext& ctx){
        ctx_print_debug(ctx);
        sift_mode_check()
        check_available_prg(BYTECODE_TRACK_ARRAY_INDEX_ID_SZ * 2)
        tai_t i_start = ctx.extract_track_array_index();
        if (i_start >= ctx.selection_array_len)
            smitsya(selection_arr_out_of_range);
        tai_t i_end = ctx.extract_track_array_index();
        if (i_end >= ctx.selection_array_len)
            smitsya(selection_arr_out_of_range);
        ctx.RAX = get_selarr_el_dist(ctx.active_thread.SAptr, i_start, i_end);
        ctx.RBX = get_selarr_el_dist(ctx.sifting_with->SAptr, i_start, i_end);
    }

    void finish_conflict_homesteader_wins(VMContext& ctx){
        ctx_print_debug(ctx);
        if (ctx.who_started_sift == opcodes::READ){
            ctx.active_thread.delete_thread();
            ctx.try_to_continue_scheduled();
        } else {
            /* FORK or MATCH (which will also be shown as FORK) */
            /* Cloning conflict ends, active_thread jumps to offsprings IP */
            ctx.active_thread.IP = ctx.child_ret_IP;
        }
        ctx.sifting_with = NULL;
    }

    void finish_conflict_intruder_wins(VMContext& ctx){
        ctx_print_debug(ctx);
        ctx.sifting_with->delete_thread();
        ctx.active_thread.IP = ctx.intruder_IP;
        if (ctx.who_started_sift == opcodes::READ){
            /* noncloning conflict won by intruder+ */
            *ctx.sifting_with = ctx.active_thread;
            ctx.active_thread.slot_occupation_status = SLOT_EMPTY_val;
            ctx.try_to_continue_scheduled();
        } else {
            /* End of cloning conflict (it involved cloning) */
            clone_thread_into_slot(ctx.active_thread, *ctx.sifting_with);
            ctx.active_thread.IP = ctx.child_ret_IP;
        }
        ctx.sifting_with = NULL;
    }

    void i_SIFTPRIOR_MIN_RABX(VMContext& ctx){
        ctx_print_debug(ctx);
        sift_mode_check()
        if (ctx.RAX < ctx.RBX){
            finish_conflict_intruder_wins(ctx);
        } else if (ctx.RAX > ctx.RBX){
            finish_conflict_homesteader_wins(ctx);
        }
    }

    void i_SIFTPRIOR_MAX_RABX(VMContext& ctx){
        ctx_print_debug(ctx);
        sift_mode_check()
        if (ctx.RAX > ctx.RBX){
            finish_conflict_intruder_wins(ctx);
        } else if (ctx.RAX < ctx.RBX){
            finish_conflict_homesteader_wins(ctx);
        }
    }

    void i_SIFT_DONE(VMContext& ctx){
        ctx_print_debug(ctx);
        sift_mode_check()
        finish_conflict_homesteader_wins(ctx);
    }

    /* Can give errors */
    void ca_branch_new_node(VMContext& ctx, tai_t key, uint64_t val){
        ctx_print_debug(ctx);
        if (ctx.CAN_total >= ctx.CA_TREE_LIMIT)
            smitsya(ca_tree_limit_violation);
        CollectionArrayNode* node = new CollectionArrayNode{key, val, ctx.active_thread.CAHptr, 1};
        // if (ctx.active_thread.CAHptr)
        // (ctx.active_thread.CAHptr->refs)++;
        ctx.active_thread.CAHptr = node;
        ctx.CAN_total++;
    }

    void i_MOV_COLARR_IMM(VMContext& ctx){
        ctx_print_debug(ctx);
        general_matching_mode_check()
        check_available_prg(BYTECODE_TRACK_ARRAY_INDEX_ID_SZ + 8)
        tai_t ca_ind = ctx.extract_track_array_index();
        uint64_t imm = ctx.extract_qw();
        ca_branch_new_node(ctx, ca_ind, imm);
    }

    void i_MOV_COLARR_BTPOS(VMContext& ctx){
        ctx_print_debug(ctx);
        general_matching_mode_check()
        check_available_prg(BYTECODE_TRACK_ARRAY_INDEX_ID_SZ)
        tai_t ca_ind = ctx.extract_track_array_index();
        ca_branch_new_node(ctx, ca_ind, ctx.passed_bytes);
    }

    /* Can throw error, should be placed at the end. Call ONLY in general matching mode */
    void edit_selection_array(VMContext& ctx, uint64_t key, uint64_t val){
        ctx_print_debug(ctx);
        uint64_t N = ctx.selection_array_len;
        if (key >= N)
            smitsya(selection_arr_out_of_range);
        if (!ctx.active_thread.SAptr){
            uint64_t* sa_instance = (uint64_t*)calloc(N + 1, 8);
            if (!sa_instance)
                throw std::bad_alloc();
            sa_instance[0] = 1;
            sa_instance[key + 1] = val;
            ctx.active_thread.SAptr = sa_instance;
        } else if (ctx.active_thread.SAptr[0] == 1){
            ctx.active_thread.SAptr[key + 1] = val;
        } else {
            uint64_t* sa_instance = (uint64_t*)calloc(N + 1, 8);
            if (!sa_instance)
                throw std::bad_alloc();
            sa_instance[0] = 1;
            for (uint64_t i = 1; i <= ctx.selection_array_len; i++)
                sa_instance[i] = ctx.active_thread.SAptr[i];
            sa_instance[key + 1] = val;
            ctx.active_thread.SAptr[0]--;
            ctx.active_thread.SAptr = sa_instance;
        }
    }

    void i_MOV_SELARR_IMM(VMContext& ctx){
        ctx_print_debug(ctx);
        general_matching_mode_check()
        check_available_prg(BYTECODE_TRACK_ARRAY_INDEX_ID_SZ + 8)
        tai_t sa_ind = ctx.extract_track_array_index();
        uint64_t imm = ctx.extract_qw();
        edit_selection_array(ctx, sa_ind, imm);
    }

    void i_MOV_SELARR_CHPOS(VMContext& ctx){
        ctx_print_debug(ctx);
        general_matching_mode_check()
        check_available_prg(BYTECODE_TRACK_ARRAY_INDEX_ID_SZ)
        tai_t sa_ind = ctx.extract_track_array_index();
        edit_selection_array(ctx, sa_ind, ctx.passed_chars);
    }

    void calloc_stack_slots(SSID_Stack& stack, sslot_id_t nmemb) {
        assert(stack.max_size == 0 && stack.sz == 0 && !stack.slots);
        sslot_id_t* storage = static_cast<sslot_id_t *>(calloc(nmemb, sizeof(sslot_id_t)));
        if (!storage)
            throw std::bad_alloc();
        stack.slots = storage;
        stack.max_size = nmemb;
    }

    Thread* calloc_slots_array(sslot_id_t nmemb) {
        Thread* ptr = static_cast<Thread *>(calloc(nmemb, sizeof(Thread)));
        if (!ptr)
            throw std::bad_alloc();
        return ptr;
    }

    void i_INIT(VMContext& ctx){
        ctx_print_debug(ctx);
        initialization_phase_check()
        if (ctx.selection_array_len > ctx.SA_LEN_LIMIT)
            smitsya(sa_length_limit_violation);
        if (ctx.read_slots_number > ctx.READ_SS_LIMIT)
            smitsya(read_sslot_count_limit_violation);
        if (ctx.fork_slots_number > ctx.FORK_SS_LIMIT)
            smitsya(fork_sslot_count_limit_violation);
        if (ctx.portion_of_second_read_halt_ns > ctx.read_slots_number)
            smitsya(fork_sslot_out_of_range);
        ctx.READ_halted_slots = calloc_slots_array(ctx.read_slots_number);
        calloc_stack_slots(ctx.READ_halted_stack_old, ctx.read_slots_number);

        ctx.portion_of_FIRST_read_halt_ns = ctx.read_slots_number - ctx.portion_of_second_read_halt_ns;
        calloc_stack_slots(ctx.READ_halted_stack_new_first, ctx.portion_of_FIRST_read_halt_ns);
        calloc_stack_slots(ctx.READ_halted_stack_new_second, ctx.portion_of_second_read_halt_ns);

        ctx.FORK_halted_slots = calloc_slots_array(ctx.fork_slots_number);
        calloc_stack_slots(ctx.FORK_halted_stack, ctx.fork_slots_number);

        ctx.initialized = true;
        ctx.unnatural_started_thread_IP = ctx.active_thread.IP;
        ctx.active_thread.delete_thread();
    }

    void i_THROW(VMContext& ctx){
        ctx.error = error_codes::program_throw;
    }

    void instruction_table(VMContext &ctx) {
        ctx_print_debug(ctx);
        uint8_t opcode = ctx.extract_instruction();

#define rcase(inst) case opcodes::inst: return i_ ## inst (ctx);
#define jumpC(UN, st) case opcodes::JC ## UN ## _B: return i_JC<st, immArgByte>(ctx); \
    case opcodes::JC ## UN ## _W: return i_JC<st, immArgWord>(ctx); \
    case opcodes::JC ## UN ## _DW: return i_JC<st, immArgDoubleWord>(ctx); \
    case opcodes::JC ## UN ## _QW: return i_JC<st, immArgQuadWord>(ctx);
        switch (opcode) {
            rcase(READ)
            rcase(READZ)
            rcase(JUMP)

            jumpC(EQUAL, condEqual)
            jumpC(LESS, condLess)
            jumpC(GRTR, condGrtr)

            rcase(FORK)
            rcase(MATCH)
            rcase(DIE)
            rcase(PARAM_READ_SS_NUMBER)
            rcase(PARAM_FORK_SS_NUMBER)
            rcase(PARAM_SELARR_LEN)
            rcase(PARAM_COLSIFTFUNC_SET)
            rcase(PARAM_COLSIFTFUNC_WIPE)
            rcase(MSG_MULTISTART_ALLOWED)
            rcase(MSG_FED_INPUT_EXTENDED)
            rcase(DMOV_RABX_SELARR)
            rcase(DDIST_RABX_SELARR)
            rcase(SIFTPRIOR_MIN_RABX)
            rcase(SIFTPRIOR_MAX_RABX)
            rcase(SIFT_DONE)
            rcase(MOV_COLARR_IMM)
            rcase(MOV_COLARR_BTPOS)
            rcase(MOV_SELARR_IMM)
            rcase(MOV_SELARR_CHPOS)
            rcase(INIT)
            rcase(THROW)
            default:
                ctx.error = error_codes::invalid_opcode;
        }
    }
}